
"""Tests for the conversions.py module."""
from datetime import datetime, timedelta
import pandas as pd
import numpy as np
import pytest
from helpsk.conversions import (
    _sort_intervals,
    cohorted_conversion_rates,
    plot_cohorted_conversion_rates,
    retention_matrix
)


def test_cohorted_conversion_rate__day_cohort(conversions):
    df = conversions.copy()
    # Define your intervals and current datetime
    intervals = [(1, 'seconds'), (1, 'minutes'), (2, 'hours'), (1, 'days'), (30, 'days')]
    current_datetime = '2023-02-02 00:00:00'

    # Create cohorts based on created_at date
    df['cohort'] = df['created_at'].dt.to_period('D').dt.to_timestamp()

    df_copy = df.copy()
    result = cohorted_conversion_rates(
        df=df,
        base_timestamp='created_at',
        conversion_timestamp='conversion_1',
        cohort='cohort',
        intervals=intervals,
        current_datetime=current_datetime,
    )
    assert df.equals(df_copy)
    assert not result['cohort'].duplicated().any()
    assert set(result['cohort']) == set(df['cohort'].unique())
    assert result['# of records'].tolist() == [3, 1, 1, 1, 1, 1]
    # all records have had a chance to convert
    # for 2023-01-01, one converts exactly at the created_at time (which doesn't count);
    # another happens before the created_at time, and one happens 1 second after)
    assert result['1 seconds'].round(3).tolist() == [0.333, 0, 0, 0, 0, 0]
    assert result['1 minutes'].round(3).tolist() == [0.333, 0, 0, 0, 0, 1]
    assert result['2 hours'].round(3).tolist() == [0.333, 0, 0, 0, 0, 1]
    assert result['1 days'].round(3).tolist() == [0.333, 0, 1, 0, 0, 1]
    assert result['30 days'].iloc[0].round(3) == 0.333
    assert result['30 days'].iloc[1] == 0
    assert pd.isna(result['30 days'].iloc[2])
    assert pd.isna(result['30 days'].iloc[3])
    assert pd.isna(result['30 days'].iloc[4])
    assert pd.isna(result['30 days'].iloc[5])

    current_datetime = '2023-01-25 23:59:50'

    # Call the conversion_rate function
    result = cohorted_conversion_rates(
        df=df,
        base_timestamp='created_at',
        conversion_timestamp='conversion_1',
        cohort='cohort',
        intervals=intervals,
        current_datetime=current_datetime,
    )
    assert df.equals(df_copy)
    assert not result['cohort'].duplicated().any()
    assert set(result['cohort']) == set(df['cohort'].unique())
    assert result['# of records'].tolist() == [3, 1, 1, 1, 1, 1]

    assert result['1 seconds'].iloc[0].round(3) == 0.333
    assert result['1 seconds'].iloc[1] == 0
    assert result['1 seconds'].iloc[2] == 0
    assert result['1 seconds'].iloc[3] == 0
    assert pd.isna(result['1 seconds'].iloc[4])
    assert pd.isna(result['1 seconds'].iloc[5])

    assert result['1 minutes'].iloc[0].round(3) == 0.333
    assert result['1 minutes'].iloc[1] == 0
    assert result['1 minutes'].iloc[2] == 0
    assert result['1 minutes'].iloc[3] == 0
    assert pd.isna(result['1 minutes'].iloc[4])
    assert pd.isna(result['1 minutes'].iloc[5])

    assert result['2 hours'].iloc[0].round(3) == 0.333
    assert result['2 hours'].iloc[1] == 0
    assert result['2 hours'].iloc[2] == 0
    assert result['2 hours'].iloc[3] == 0
    assert pd.isna(result['2 hours'].iloc[4])
    assert pd.isna(result['2 hours'].iloc[5])

    assert result['1 days'].iloc[0].round(3) == 0.333
    assert result['1 days'].iloc[1] == 0
    assert result['1 days'].iloc[2] == 1
    assert pd.isna(result['1 days'].iloc[3])
    assert pd.isna(result['1 days'].iloc[4])
    assert pd.isna(result['1 days'].iloc[5])

    assert pd.isna(result['30 days'].iloc[0])
    assert pd.isna(result['30 days'].iloc[1])
    assert pd.isna(result['30 days'].iloc[2])
    assert pd.isna(result['30 days'].iloc[3])
    assert pd.isna(result['30 days'].iloc[4])
    assert pd.isna(result['30 days'].iloc[5])


def test_cohorted_conversion_rate__week_cohort(conversions):
    df = conversions.copy()
    # Define your intervals and current datetime
    intervals = [(1, 'seconds'), (1, 'minutes'), (2, 'hours'), (1, 'days'), (30, 'days')]
    current_datetime = '2023-02-02 00:00:00'

    # Create cohorts based on created_at date
    df['cohort'] = df['created_at'].dt.to_period('W').dt.to_timestamp()

    df_copy = df.copy()
    result = cohorted_conversion_rates(
        df=df,
        base_timestamp='created_at',
        conversion_timestamp='conversion_1',
        cohort='cohort',
        intervals=intervals,
        current_datetime=current_datetime,
    )
    assert df.equals(df_copy)
    assert not result['cohort'].duplicated().any()
    assert set(result['cohort']) == set(df['cohort'].unique())
    assert result['# of records'].tolist() == [3, 1, 1, 1, 2]
    # all records have had a chance to convert
    # for 2023-01-01, one converts exactly at the created_at time (which doesn't count);
    # another happens before the created_at time, and one happens 1 second after)
    assert result['1 seconds'].round(3).tolist() == [0.333, 0, 0, 0, 0]
    assert result['1 minutes'].round(3).tolist() == [0.333, 0, 0, 0, 0.5]
    assert result['2 hours'].round(3).tolist() == [0.333, 0, 0, 0, 0.5]
    assert result['1 days'].round(3).tolist() == [0.333, 0, 1, 0, 0.5]
    assert result['30 days'].iloc[0].round(3) == 0.333
    assert result['30 days'].iloc[1] == 0
    assert pd.isna(result['30 days'].iloc[2])
    assert pd.isna(result['30 days'].iloc[3])
    assert pd.isna(result['30 days'].iloc[4])

    current_datetime = '2023-01-25 23:59:50'

    # Call the conversion_rate function
    result = cohorted_conversion_rates(
        df=df,
        base_timestamp='created_at',
        conversion_timestamp='conversion_1',
        cohort='cohort',
        intervals=intervals,
        current_datetime=current_datetime,
    )
    assert df.equals(df_copy)
    assert not result['cohort'].duplicated().any()
    assert set(result['cohort']) == set(df['cohort'].unique())
    assert result['# of records'].tolist() == [3, 1, 1, 1, 2]

    assert result['1 seconds'].iloc[0].round(3) == 0.333
    assert result['1 seconds'].iloc[1] == 0
    assert result['1 seconds'].iloc[2] == 0
    assert result['1 seconds'].iloc[3] == 0
    assert pd.isna(result['1 seconds'].iloc[4])

    assert result['1 minutes'].iloc[0].round(3) == 0.333
    assert result['1 minutes'].iloc[1] == 0
    assert result['1 minutes'].iloc[2] == 0
    assert result['1 minutes'].iloc[3] == 0
    assert pd.isna(result['1 minutes'].iloc[4])

    assert result['2 hours'].iloc[0].round(3) == 0.333
    assert result['2 hours'].iloc[1] == 0
    assert result['2 hours'].iloc[2] == 0
    assert result['2 hours'].iloc[3] == 0
    assert pd.isna(result['2 hours'].iloc[4])

    assert result['1 days'].iloc[0].round(3) == 0.333
    assert result['1 days'].iloc[1] == 0
    assert result['1 days'].iloc[2] == 1
    assert pd.isna(result['1 days'].iloc[3])
    assert pd.isna(result['1 days'].iloc[4])

    assert pd.isna(result['30 days'].iloc[0])
    assert pd.isna(result['30 days'].iloc[1])
    assert pd.isna(result['30 days'].iloc[2])
    assert pd.isna(result['30 days'].iloc[3])
    assert pd.isna(result['30 days'].iloc[4])


def test_cohorted_conversion_rate__groups(conversions):
    df = conversions.copy()
    # Define your intervals and current datetime
    intervals = [(1, 'seconds'), (1, 'minutes'), (2, 'hours'), (1, 'days'), (30, 'days')]
    current_datetime = '2023-01-25 23:59:50'
    df['cohort'] = df['created_at'].dt.to_period('W').dt.to_timestamp()

    df_copy = df.copy()
    result = cohorted_conversion_rates(
        df=df,
        base_timestamp='created_at',
        conversion_timestamp='conversion_1',
        cohort='cohort',
        intervals=intervals,
        groups='segments',
        current_datetime=current_datetime,
    )
    assert df.equals(df_copy)
    assert not result[['cohort', 'segments']].duplicated().any()
    assert set(result['cohort']) == set(df['cohort'].unique())
    assert result['# of records'].tolist() == [2, 1, 1, 1, 1, 2]

    assert result['1 seconds'].iloc[0] == 0.5
    assert result['1 seconds'].iloc[1] == 0
    assert result['1 seconds'].iloc[2] == 0
    assert result['1 seconds'].iloc[3] == 0
    assert result['1 seconds'].iloc[4] == 0
    assert pd.isna(result['1 seconds'].iloc[5])

    assert result['1 minutes'].iloc[0] == 0.5
    assert result['1 minutes'].iloc[1] == 0
    assert result['1 minutes'].iloc[2] == 0
    assert result['1 minutes'].iloc[3] == 0
    assert result['1 minutes'].iloc[4] == 0
    assert pd.isna(result['1 minutes'].iloc[5])

    assert result['2 hours'].iloc[0] == 0.5
    assert result['2 hours'].iloc[1] == 0
    assert result['2 hours'].iloc[2] == 0
    assert result['2 hours'].iloc[3] == 0
    assert result['2 hours'].iloc[4] == 0
    assert pd.isna(result['2 hours'].iloc[5])

    assert result['1 days'].iloc[0] == 0.5
    assert result['1 days'].iloc[1] == 0
    assert result['1 days'].iloc[2] == 0
    assert result['1 days'].iloc[3] == 1
    assert pd.isna(result['1 days'].iloc[4])
    assert pd.isna(result['1 days'].iloc[5])

    assert pd.isna(result['30 days'].iloc[0])
    assert pd.isna(result['30 days'].iloc[1])
    assert pd.isna(result['30 days'].iloc[2])
    assert pd.isna(result['30 days'].iloc[3])
    assert pd.isna(result['30 days'].iloc[4])
    assert pd.isna(result['30 days'].iloc[5])


def test_plot_cohorted_conversion_rates(conversions):
    df = conversions.copy()
    df['cohort'] = df['created_at'].dt.to_period('W').dt.to_timestamp()
    _ = plot_cohorted_conversion_rates(
        df=df,
        base_timestamp='created_at',
        conversion_timestamp='conversion_1',
        cohort='cohort',
        intervals=[(1, 'seconds'), (1, 'minutes'), (2, 'hours'), (1, 'days'), (30, 'days')],
        groups=None,
        current_datetime='2023-01-25 23:59:50',
    )
    _ = plot_cohorted_conversion_rates(
        df=df,
        base_timestamp='created_at',
        conversion_timestamp='conversion_1',
        cohort='cohort',
        intervals=[(1, 'seconds'), (1, 'minutes'), (2, 'hours'), (1, 'days'), (30, 'days')],
        groups='segments',
        category_orders={},
        current_datetime='2023-01-25 23:59:50',
    )
    _ = plot_cohorted_conversion_rates(
        df=df,
        base_timestamp='created_at',
        conversion_timestamp='conversion_1',
        cohort='cohort',
        intervals=[(1, 'seconds'), (1, 'minutes'), (2, 'hours'), (1, 'days'), (30, 'days')],
        groups='segments',
        current_datetime='2023-01-25 23:59:50',
        graph_type='bar',
        title='Title',
        subtitle='Subtitle',
        x_axis_label='X Axis Label',
        y_axis_label='Y Axis Label',
        legend_label='Legend Label',
        facet_col_wrap=6,
        bar_mode='relative',
        category_orders={'abc': ['a', 'b', 'c']},
        opacity=0.9,
        height=800,
        width=100,
        free_y_axis=False,
    )
    _ = plot_cohorted_conversion_rates(
        df=df,
        base_timestamp='created_at',
        conversion_timestamp='conversion_1',
        cohort='cohort',
        intervals=[(1, 'seconds'), (1, 'minutes'), (2, 'hours'), (1, 'days'), (30, 'days')],
        groups='segments',
        current_datetime='2023-01-25 23:59:50',
        graph_type='line',
        title='Title',
        subtitle='Subtitle',
        x_axis_label='X Axis Label',
        y_axis_label='Y Axis Label',
        legend_label='Legend Label',
        facet_col_wrap=6,
        bar_mode='relative',
        opacity=0.9,
        height=800,
        width=100,
        free_y_axis=True,
    )


def test__sort_intervals():
    intervals = [
        (120, 'seconds'),
        (30, 'minutes'),
        (2, 'hours'),
        (1, 'days'),
    ]
    sorted_intervals = _sort_intervals(intervals)
    assert sorted_intervals == [
        (1, 'days'),
        (2, 'hours'),
        (30, 'minutes'),
        (120, 'seconds'),
    ]

    intervals = [
        (45, 'seconds'),
        (1, 'hours'),
        (5, 'minutes'),
    ]
    sorted_intervals = _sort_intervals(intervals)
    assert sorted_intervals == [
        (1, 'hours'),
        (5, 'minutes'),
        (45, 'seconds'),
    ]

    intervals = [
        (1, 'days'),
        (14, 'days'),
        (7, 'days'),
    ]
    sorted_intervals = _sort_intervals(intervals)
    assert sorted_intervals == [
        (14, 'days'),
        (7, 'days'),
        (1, 'days'),
    ]


def generate_fake_data():
    base_date = datetime(2023, 1, 1)
    data = []
    for user_id in range(100):
        for days in range(0, 70, 1):
            event_datetime = base_date + timedelta(days=days + user_id)
            if event_datetime > base_date + timedelta(days=100):
                continue
            if np.random.rand() < 0.5:
                # print('append', user_id, days)
                data.append({
                    'user_id': user_id,
                    'datetime': event_datetime
                })
            if np.random.rand() < 0.5:
                # print('append', user_id, days)
                data.append({
                    'user_id': user_id,
                    'datetime': event_datetime
                })
    return pd.DataFrame(data, columns=['user_id', 'datetime'])


def test_retention_matrix():  # noqa
    df = generate_fake_data()
    copy_df = df.copy()
    retention_week = retention_matrix(
        df,
        timestamp='datetime',
        unique_id='user_id',
        intervals='week',
        current_datetime=df['datetime'].max(),
    )
    assert df.equals(copy_df)
    assert retention_week['cohort'].notna().all()
    assert retention_week['# of unique ids'].notna().all()
    assert (retention_week.drop(columns=['cohort', '# of unique ids']).fillna(0) <= 1).all().all()
    assert (retention_week['# of unique ids'] > 1).any()
    assert (retention_week['0'] == 1).all().all()
    assert retention_week['# of unique ids'].sum() == df['user_id'].nunique()
    expected_cohort_sizes = (
        df
        .groupby('user_id')
        .agg(min_date=('datetime', 'min'))
        .assign(cohort=lambda x: x['min_date'].dt.to_period('W').dt.to_timestamp())
        .groupby('cohort')
        .size()
        .sort_index()
        .values
        .tolist()
    )
    assert retention_week['# of unique ids'].tolist() == expected_cohort_sizes

    retention_day = retention_matrix(
        df,
        timestamp='datetime',
        unique_id='user_id',
        intervals='day',
    )
    assert df.equals(copy_df)
    assert retention_day['cohort'].notna().all()
    assert retention_day['# of unique ids'].notna().all()
    assert (retention_day.drop(columns=['cohort', '# of unique ids']).fillna(0) <= 1).all().all()
    assert (retention_day['# of unique ids'] > 1).any()
    assert (retention_day['0'] == 1).all().all()
    assert retention_day['# of unique ids'].sum() == df['user_id'].nunique()
    expected_cohort_sizes = (
        df
        .groupby('user_id')
        .agg(min_date=('datetime', 'min'))
        .assign(cohort=lambda x: x['min_date'].dt.to_period('D').dt.to_timestamp())
        .groupby('cohort')
        .size()
        .sort_index()
        .values
        .tolist()
    )
    assert retention_day['# of unique ids'].tolist() == expected_cohort_sizes

    # test with min_events == 2
    retention2 = retention_matrix(
        df,
        timestamp='datetime',
        unique_id='user_id',
        intervals='week',
        current_datetime=df['datetime'].max(),
        min_events=2,
    )
    assert retention2['cohort'].equals(retention_week['cohort'])
    assert retention2['# of unique ids'].equals(retention_week['# of unique ids'])
    assert retention2['0'].equals(retention_week['0'])
    _retention_2 = retention2.drop(columns=['cohort', '# of unique ids', '0']).fillna(0)
    _retention_week = retention_week.drop(columns=['cohort', '# of unique ids', '0']).fillna(0)
    assert (_retention_2 < _retention_week).any().any()
    assert retention2['cohort'].notna().all()
    assert retention2['# of unique ids'].notna().all()
    assert (retention2.drop(columns=['cohort', '# of unique ids']).fillna(0) <= 1).all().all()
    assert (retention2['# of unique ids'] > 1).any()
    assert (retention2['0'] == 1).all().all()
    assert retention2['# of unique ids'].sum() == df['user_id'].nunique()
    expected_cohort_sizes = (
        df
        .groupby('user_id')
        .agg(min_date=('datetime', 'min'))
        .assign(cohort=lambda x: x['min_date'].dt.to_period('W').dt.to_timestamp())
        .groupby('cohort')
        .size()
        .sort_index()
        .values
        .tolist()
    )
    assert retention2['# of unique ids'].tolist() == expected_cohort_sizes

    df.iloc[0, 0] = np.nan
    with pytest.raises(AssertionError):
        retention_matrix(
            df,
            timestamp='datetime',
            unique_id='user_id',
            intervals='week',
        )


def test_retention_matrix_by_month():  # noqa
    df = generate_fake_data()
    copy_df = df.copy()
    retention_month = retention_matrix(
        df,
        timestamp='datetime',
        unique_id='user_id',
        intervals='month',
        current_datetime=df['datetime'].max(),
    )
    assert df.equals(copy_df)
    assert retention_month['cohort'].notna().all()
    assert retention_month['# of unique ids'].notna().all()
    assert (retention_month.drop(columns=['cohort', '# of unique ids']).fillna(0) <= 1).all().all()
    assert (retention_month['# of unique ids'] > 1).any()
    assert (retention_month['0'] == 1).all().all()
    assert retention_month['# of unique ids'].sum() == df['user_id'].nunique()
    expected_cohort_sizes = (
        df
        .groupby('user_id')
        .agg(min_date=('datetime', 'min'))
        .assign(cohort=lambda x: x['min_date'].dt.to_period('M').dt.to_timestamp())
        .groupby('cohort')
        .size()
        .sort_index()
        .values
        .tolist()
    )
    assert retention_month['# of unique ids'].tolist() == expected_cohort_sizes

    retention_month_28 = retention_matrix(
        df,
        timestamp='datetime',
        unique_id='user_id',
        intervals='month',
        current_datetime=df['datetime'].max(),
        min_events=5,
    )
    # this could randomly fail if we don't have any retained users in the last month
    assert retention_month.columns.tolist() == retention_month_28.columns.tolist()
    assert df.equals(copy_df)
    assert retention_month_28['cohort'].notna().all()
    assert retention_month_28['# of unique ids'].notna().all()
    assert (retention_month_28.drop(columns=['cohort', '# of unique ids']).fillna(0) <= 1).all().all()
    assert (retention_month_28['# of unique ids'] > 1).any()
    assert (retention_month_28['0'] == 1).all().all()
    assert retention_month_28['# of unique ids'].sum() == df['user_id'].nunique()
    expected_cohort_sizes = (
        df
        .groupby('user_id')
        .agg(min_date=('datetime', 'min'))
        .assign(cohort=lambda x: x['min_date'].dt.to_period('M').dt.to_timestamp())
        .groupby('cohort')
        .size()
        .sort_index()
        .values
        .tolist()
    )
    assert retention_month_28['# of unique ids'].tolist() == expected_cohort_sizes
    _retention_month = retention_month.drop(columns=['cohort', '# of unique ids', '0']).fillna(0)
    _retention_month_28 = retention_month_28.drop(columns=['cohort', '# of unique ids', '0']).fillna(0)
    assert (_retention_month > _retention_month_28).any().any()


def test_retention__duplicate_user_cohort_bug():  # noqa
    df = pd.DataFrame({
        'categories': pd.Categorical(['a', 'b', 'c', 'a', 'b']),
        'dates': pd.to_datetime(['2023-01-01', '2023-01-02', '2023-01-03', '2023-01-04', '2023-01-05']),  # noqa
        'datetimes': pd.to_datetime(['2023-01-01 01:01:01', '2023-01-02 02:02:02', '2023-01-03 03:03:03', '2023-01-04 04:04:04', '2023-01-05 05:05:05']),  # noqa
    })
    copy_df = df.copy()
    retention = retention_matrix(
        df,
        timestamp='datetimes',
        unique_id='categories',
        intervals='week',
        current_datetime=None,
    )
    assert df.equals(copy_df)
    assert retention['cohort'].notna().all()
    assert retention['# of unique ids'].notna().all()
    assert (retention.drop(columns=['cohort', '# of unique ids']).fillna(0) <= 1).all().all()
    assert (retention['# of unique ids'] > 1).any()
    assert (retention['0'] == 1).all().all()
    assert retention['# of unique ids'].sum() == df['categories'].nunique()
    expected_cohort_sizes = (
        df
        .groupby('categories')
        .agg(min_date=('datetimes', 'min'))
        .assign(cohort=lambda x: x['min_date'].dt.to_period('W').dt.to_timestamp())
        .groupby('cohort')
        .size()
        .sort_index()
        .values
        .tolist()
    )
    assert retention['# of unique ids'].tolist() == expected_cohort_sizes

    retention = retention_matrix(
        df,
        timestamp='datetimes',
        unique_id='categories',
        intervals='week',
        current_datetime=None,
        min_events=2,
    )
    assert df.equals(copy_df)
    assert retention['cohort'].notna().all()
    assert retention['# of unique ids'].notna().all()
    assert (retention.drop(columns=['cohort', '# of unique ids']).fillna(0) <= 1).all().all()
    assert (retention['# of unique ids'] > 1).any()
    assert (retention['0'] == 1).all().all()
    assert retention['# of unique ids'].sum() == df['categories'].nunique()
    expected_cohort_sizes = (
        df
        .groupby('categories')
        .agg(min_date=('datetimes', 'min'))
        .assign(cohort=lambda x: x['min_date'].dt.to_period('W').dt.to_timestamp())
        .groupby('cohort')
        .size()
        .sort_index()
        .values
        .tolist()
    )
    assert retention['# of unique ids'].tolist() == expected_cohort_sizes



"""Contains helper functions for plotting."""

from __future__ import annotations

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

import helpsk.pandas as hpandas
from helpsk.exceptions import HelpskParamValueError

STANDARD_WIDTH = 10
GOLDEN_RATIO = 1.61803398875
STANDARD_HEIGHT = STANDARD_WIDTH / GOLDEN_RATIO
STANDARD_WIDTH_HEIGHT = (STANDARD_WIDTH, STANDARD_HEIGHT)


def plot_value_frequency(
        series: pd.Series, sort_by_frequency: bool = True,
        figure_size: tuple[int, int] = STANDARD_WIDTH_HEIGHT,
        x_axis_rotation: int = 30) -> None:
    """
    Shows the unique values and corresponding frequencies.

    Args:
        series:
            a Pandas series, either categorical or with integers.
        sort_by_frequency:
            if True then sort by frequency desc; otherwise sort by index (either numerically
            ascending if series is numeric, or alphabetically if non-ordered categoric, or by
            category if ordered categoric
        figure_size:
            tuple containing `(width, height)` of plot. The default height is defined by
            `STANDARD_HEIGHT`, and the default width is `STANDARD_HEIGHT / GOLDEN_RATIO`
        x_axis_rotation:
            the angle to rotate the x-axis text.
    """
    value_frequencies = hpandas.value_frequency(series=series, sort_by_frequency=sort_by_frequency)
    plot_object = value_frequencies.\
        drop(columns='Percent').\
        plot(
            kind='bar', rot=10,
            ylabel='Frequency',
            title=f'Name: `{series.name}`',
            legend=False,
            figsize=figure_size,
        )
    for idx, label in enumerate(list(value_frequencies.index)):
        if pd.isna(label):
            frequency = value_frequencies[value_frequencies.index.isna()]['Frequency'].iloc[0]
            percent = value_frequencies[value_frequencies.index.isna()]['Percent'].iloc[0]
        else:
            frequency = value_frequencies.loc[label, 'Frequency']
            percent = value_frequencies.loc[label, 'Percent']
        plot_object.annotate(frequency,
                             xy=(idx, frequency),
                             xytext=(0, 2),
                             ha='center',
                             textcoords='offset points')
        plot_object.annotate(f"{percent * 100:.0f}%",
                             xy=(idx, max(frequency - 10, 2)),
                             xytext=(0, 0),
                             color='white',
                             ha='center',
                             textcoords='offset points')

    plot_object.set_xticklabels(
        labels=value_frequencies.index.values, rotation=x_axis_rotation, ha='right',
    )
    plt.tight_layout()


def plot_correlation_heatmap(
        dataframe: pd.DataFrame,
        threshold: float | None = None,
        title: str | None = None,
        figure_size: tuple[int, int] = STANDARD_WIDTH_HEIGHT,
        round_by: int = 2,
        features_to_highlight: list[str] | None = None) -> None:
    """
    Creates a heatmap of the correlations between all of the numeric columns.

    Args:
        dataframe:
            dataframe to get correlations from. Automatically chooses all numeric columns.
        threshold:
            threshold: the heatmap only includes columns that have a correlation value,
            corresponding to at least one other column, where the absolute value is higher than the
            threshold.

        So for example, if the threshold is `0.8` then all columns that have a correlation
            (absolute) value of `>=.80` anywhere in the correlation matrix (i.e. with any other
            column), will show in the heatmap, even though a specific column might not have high
            correlations with every other column included.
        title:
            title of plot
        figure_size:
            tuple containing `(width, height)` of plot. The default height is defined by
            `STANDARD_HEIGHT`, and the default width is `STANDARD_HEIGHT / GOLDEN_RATIO`
        round_by:
            the number of decimal places to round to when showing the correlations in the heatmap
        features_to_highlight:
            feature labels to highlight in red
    """
    correlations = dataframe.corr(numeric_only=True)

    if threshold is not None:
        features = correlations.columns.to_numpy()
        correlation_matrix = np.abs(correlations.values)
        np.fill_diagonal(correlation_matrix, np.NaN)
        meets_threshold = np.apply_along_axis(
            lambda x: np.any(x >= threshold),
            0,
            correlation_matrix,
        )
        if not meets_threshold.any():
            raise HelpskParamValueError('correlation `threshold` set too high.')

        features = features[meets_threshold]
        correlations = correlations.loc[features, features]

    if title is None:
        title = ''

    # remove duplicates (i.e. mirror image in second half of heatmap
    mask = np.zeros_like(correlations)
    mask[np.triu_indices_from(mask)] = True

    with sns.axes_style("white"):
        figure, axis = plt.subplots()
        figure.set_size_inches(figure_size[0], figure_size[1])
        sns.heatmap(correlations,
                    mask=mask,
                    annot=True,
                    fmt=f'.{round_by}f',
                    cmap=sns.diverging_palette(220, 10, as_cmap=True),
                    vmin=-1,
                    vmax=1,
                    square=True,
                    ax=axis,
                    center=0)
        plt.xticks(rotation=20, ha='right')
        plt.yticks(rotation=0)
        plt.title(title)
        plt.tight_layout()

        if features_to_highlight is not None:
            indexes_to_highlight = [index for index, x in enumerate(correlations.index.values)
                                    if x in features_to_highlight]

            for index_to_highlight in indexes_to_highlight:
                plt.gca().get_xticklabels()[index_to_highlight].set_color('red')
                plt.gca().get_yticklabels()[index_to_highlight].set_color('red')

    plt.tight_layout()


def plot_dodged_barchart(
        dataframe: pd.DataFrame,
        outer_column: str,
        inner_column: str,
        figure_size: tuple[int, int] = STANDARD_WIDTH_HEIGHT,
        missing_value_replacement: str = '<Missing>') -> None:
    """
    First attempt at dodged barchart. It needs some work.

    Args:
        dataframe:
            pandas data.frame
        outer_column:
            column name associated with outer bars
        inner_column:
            column name associated with inner bars
        figure_size:
            tuple containing `(width, height)` of plot. The default height is defined by
            `STANDARD_HEIGHT`, and the default width is `STANDARD_HEIGHT / GOLDEN_RATIO`
        missing_value_replacement:
            the value to replace missing values with
    """
    dataframe = dataframe[[outer_column, inner_column]].copy()

    # Replace all Boolean values with string because boolean causes issues
    dataframe[outer_column] = hpandas.replace_all_bools_with_strings(dataframe[outer_column])
    dataframe[inner_column] = hpandas.replace_all_bools_with_strings(dataframe[inner_column])

    dataframe[outer_column] = hpandas.fill_na(
        series=dataframe[outer_column],
        missing_value_replacement=missing_value_replacement,
    )
    dataframe[inner_column] = hpandas.fill_na(
        series=dataframe[inner_column],
        missing_value_replacement=missing_value_replacement,
    )

    outer_labels = dataframe[outer_column].unique().tolist()
    if hpandas.is_series_categorical(dataframe[outer_column]):
        categories = list(dataframe[outer_column].cat.categories)
        outer_labels = [x for x in categories if x in outer_labels]

    inner_labels = dataframe[inner_column].unique().tolist()
    if hpandas.is_series_categorical(dataframe[inner_column]):
        categories = list(dataframe[inner_column].cat.categories)
        inner_labels = [x for x in categories if x in inner_labels]

    grouped_data = dataframe.groupby([outer_column, inner_column]).size()

    outer_totals = [grouped_data[index].sum() for index in outer_labels]
    group_locations = np.arange(len(outer_labels))  # the x locations for the groups

    bar_midpoints = group_locations  # + (width / len(inner_labels))

    fig, axis = plt.subplots()
    fig.set_size_inches(figure_size[0], figure_size[1])

    outer_bar_width = 0.9
    inner_bar_width = outer_bar_width / len(inner_labels)

    ax_totals = axis.bar(x=bar_midpoints,
                         height=outer_totals,
                         width=outer_bar_width,
                         color='black', alpha=0.15)

    # from matplotlib.pyplot import cm
    # colors = cm.rainbow(np.linspace(0, 1, 10)[::-1])
    colors = ['b', 'g', 'r', 'c', 'm', 'y', 'k'] * 100

    if len(inner_labels) > len(colors):
        raise NotImplementedError(
            'Need to update implementation to use >' + str(len(colors)) + ' colors :(')

    ax_list = []
    for index in range(len(inner_labels)):
        # the 'if' is because it's not guaranteed that every class of the target variable will be
        # found in each category of 'outer_column', especially if the category is very small in
        # size.
        counts = [
            grouped_data[x, inner_labels[index]]
            if inner_labels[index] in grouped_data.loc[x].index else 0
            for x in outer_labels
        ]

        # `starting_offset` is the amount we have to subtract from our inner bars so the start of
        # the inner/outer are aligned
        starting_offset = (outer_bar_width / 2) - (inner_bar_width / 2)
        # (inner_bar_width * index) is the offset for each of the inner bars
        # i.e. it's the distance we have to add so that we aren't overlapping inner bars
        ax_list.append(
            axis.bar(x=group_locations - starting_offset + (inner_bar_width * index),
                     height=tuple(counts),
                     width=inner_bar_width,
                     color=colors[index]))

    # add some text for labels, title and axes ticks
    axis.set_ylabel('Count')
    axis.set_xlabel(outer_column)
    axis.set_title(f'`{outer_column}` vs. `{inner_column}`')
    axis.set_xticks(group_locations)
    axis.set_xticklabels(labels=outer_labels, rotation=20, ha='right')
    axis.legend(
        [ax[0] for ax in ax_list] + [ax_totals[0]], inner_labels + ['Total'],  # noqa: RUF005
        title=inner_column,
    )
    plt.tight_layout()


def plot_histogram_with_categorical(
        dataframe: pd.DataFrame,
        numeric_column: str,
        categorical_column: str,
        missing_value_replacement: str = '<Missing>') -> None:
    """
    Plots a categorical histogram within numeric histogram.

    Args:
        dataframe:
            TBD
        numeric_column:
            TBD
        categorical_column:
            TBD
        missing_value_replacement:
            the value
    """
    cut_dataframe = pd.DataFrame({
        numeric_column: pd.cut(
            dataframe[numeric_column].values,
            bins=10,
            right=True,
            include_lowest=True,
        ),
    })

    cut_dataframe[numeric_column] = hpandas.fill_na(
        cut_dataframe[numeric_column],
        missing_value_replacement=missing_value_replacement,
    )
    categories = [str(x) for x in cut_dataframe[numeric_column].cat.categories]
    cut_dataframe[numeric_column] = [str(x) for x in cut_dataframe[numeric_column]]
    cut_dataframe[numeric_column] = cut_dataframe[numeric_column].astype('category')
    # only keep the categories that appear in the data, otherwise we get an error when calling
    # reorder_categories with extra categories
    categories = [x for x in categories if x in cut_dataframe[numeric_column].cat.categories]
    cut_dataframe[numeric_column] = cut_dataframe[numeric_column].\
        cat.\
        reorder_categories(categories)

    cut_dataframe[categorical_column] = dataframe[categorical_column].copy()

    plot_dodged_barchart(
        dataframe=cut_dataframe,
        outer_column=numeric_column,
        inner_column=categorical_column,
    )



import os
import pytest
from sqlalchemy.exc import OperationalError

from helpsk.database import Sqlite
from helpsk.validation import dataframes_match

import os.path


def test_sqlite():
    import pandas as pd
    df = pd.DataFrame({
        'column_1': ['a', None, 'c'],
        'column_2': [1, 2, None],
    })

    db_path = 'tests/test.db'
    try:
        assert not os.path.exists(db_path)
        with Sqlite(path=db_path) as db:
            assert os.path.exists(db_path)

            with pytest.raises(OperationalError):
                db.query("SELECT * FROM TEST_TABLE")

            db.insert_records(
                dataframe=df,
                table='TEST_TABLE',
                create_table=False,
                overwrite=False,
            )
            test_table_df = db.query("SELECT * FROM TEST_TABLE")
            assert dataframes_match([df, test_table_df])

            db.insert_records(
                dataframe=df,
                table='TEST_TABLE',
                create_table=False,
                overwrite=True,
            )
            test_table_df = db.query("SELECT * FROM TEST_TABLE")
            assert dataframes_match([df, test_table_df])

            with pytest.raises(ValueError):
                db.insert_records(
                    dataframe=df,
                    table='TEST_TABLE',
                    create_table=True,
                    overwrite=False,
                )
            expected_df = pd.concat([df, df]).reset_index(drop=True)
            db.insert_records(dataframe=df, table='TEST_TABLE')
            test_table_df = db.query("SELECT * FROM TEST_TABLE")
            assert dataframes_match([expected_df, test_table_df])

        with Sqlite(path=db_path) as db:
            test_table_df = db.query("SELECT * FROM TEST_TABLE")
            assert dataframes_match([expected_df, test_table_df])

            db.execute_statement("DELETE FROM TEST_TABLE WHERE column_1 = 'a'")
            test_table_df = db.query("SELECT * FROM TEST_TABLE")
            assert dataframes_match([
                test_table_df,
                expected_df[expected_df['column_1'] != 'a'].reset_index(drop=True)
            ])

            db.execute_statement("DROP TABLE TEST_TABLE;")
            with pytest.raises(OperationalError):
                db.query("SELECT * FROM TEST_TABLE")
    finally:
        os.remove(db_path)



"""Collection of helpers methods that help to style tables in Jupyter Notebooks."""

from __future__ import annotations
from typing import TypeVar
from html import escape

import pandas as pd
import numpy as np

from pandas.io.formats.style import Styler
from pandas.api.types import is_list_like
from pandas._typing import Axis
from seaborn import color_palette

from helpsk import color


import helpsk.pandas as hpandas
from helpsk.validation import any_none_nan


Subset = TypeVar('Subset')


def add_bar(
        df_style: Styler,
        column_name: str,
        min_value: float | None = None,
        max_value: float | None = None,
        round_by: int = 0,
        color: str = color.GRAY) -> Styler:
    """
    Add bars to a particular column with the ability to specify a min/max value.

    Args:
        df_style: style object e.g. df.style
        column_name: name of the column to apply bars
        min_value: min value of bars
        max_value: max value of bars
        round_by: number of decimals to round by
        color: color of bars
    """
    return df_style.\
        bar(subset=[column_name], color=color, vmin=min_value, vmax=max_value).\
        format(subset=[column_name], precision=round_by)


def add_background_gradient(
        df_style: Styler,
        column_name: str,
        min_value: float | None = None,
        max_value: float | None = None,
        round_by: int = 0,
        color_cmap: str | None = None) -> Styler:
    """
    Add gradient to a particular column with the ability to specify a min/max value.

    Args:
        df_style: style object e.g. df.style
        column_name: name of the column to apply gradient
        min_value: min value of gradient
        max_value: max value of gradient
        round_by: number of decimals to round by
        color_cmap: cmap of gradient (sns.color_palette("vlag", as_cmap=True) is default)
    """
    if not color_cmap:
        color_cmap = color_palette("vlag", as_cmap=True)

    return df_style.\
        background_gradient(
            subset=[column_name],
            cmap=color_cmap,
            vmin=min_value,
            vmax=max_value,
        ).\
        format(subset=[column_name], precision=round_by)


def format(styler: pd.DataFrame | Styler,  # noqa: A001
           subset: list[str] | None = None,
           round_by: int = 2,
           fill_missing_value: str | None = '<NA>',
           missing_color: str | None = color.WARNING,
           thousands: str | None = ',',
           hide_index: bool = False) -> Styler:
    """
    Applies basic formatting to pandas Dataframe.

    Args:
        styler:
            either pd.Dataframe or pd.Dataframe.style
        subset:
            A valid 2d input to DataFrame.loc[<subset>], or, in the case of a 1d input or single
            key, to DataFrame.loc[:, <subset>] where the columns are prioritised, to limit data to
            before applying the function.
        round_by:
            number of digits to round numeric columns to
        fill_missing_value:
            the value to replace missing data (e.g. NaN)
        missing_color:
            The background color for cells that have missing values.
        thousands:
            the separator used for thousands e.g. `'` will result in `10,000` while ` ` will
            result in `10 000`.
        hide_index:
            Hide the index of the dataframe.

    Returns:
        styler
    """
    if isinstance(styler, pd.DataFrame):
        styler = styler.style

    styler = styler.highlight_null(missing_color)

    if hide_index:
        styler = styler.hide(axis='index')

    return styler.format(
        subset=subset,
        precision=round_by,
        na_rep=escape(fill_missing_value),
        thousands=thousands,
    )


def background_color(
        styler: pd.DataFrame | Styler,
        palette: str = 'Blues',
        **kwargs: dict) -> Styler:
    """
    Applies a background color to pandas Dataframe.

    Args:
        styler:
            either pd.Dataframe or pd.Dataframe.style
        palette:
            name of the palette (value passed into seaborn `color_palette()`
        kwargs:
            additional arguments that will be passed to the pandas `.background_gradient()`
            function.

    Returns:
        styler
    """
    if isinstance(styler, pd.DataFrame):
        styler = styler.style

    # color_map = sns.light_palette("green", as_cmap=True)
    # color_map = sns.color_palette("dark:salmon_r", as_cmap=True)
    # color_map = sns.color_palette(['red', 'blue', 'green'], as_cmap=True)
    # color_map = sns.color_palette("light:#5A9", as_cmap=True)
    color_map = color_palette(palette, as_cmap=True)
    return styler.background_gradient(cmap=color_map, **kwargs)


def __bar_inverse(
        style_object: Styler,
        align: str,
        colors: list[str],
        width: float = 100,
        min_value: float | None = None,
        max_value: float | None = None) -> pd.DataFrame:
    """
    CODE MODIFIED FROM FOLLOWING.

        https://github.com/pandas-dev/pandas/blob/v1.3.2/pandas/io/formats/style.py#L2178-L2258

    Draw bar chart in dataframe cells.
    """
    # Get input value range.
    object_min = np.nanmin(style_object.to_numpy()) if min_value is None else min_value
    object_max = np.nanmax(style_object.to_numpy()) if max_value is None else max_value
    if align == "mid":
        object_min = min(0, object_min)
        object_max = max(0, object_max)
    elif align == "zero":
        # For "zero" mode, we want the range to be symmetrical around zero.
        object_max = max(abs(object_min), abs(object_max))
        object_min = -object_max
    # Transform to percent-range of linear-gradient
    normed = width \
        * (style_object.to_numpy(dtype=float) - object_min) \
        / (object_max - object_min + 1e-12)
    zero = -width * object_min / (object_max - object_min + 1e-12)

    def css_bar(start: float, end: float, color: str) -> str:
        """Generate CSS code to draw a bar from start to end."""
        css = "width: 10em; height: 80%;"
        if end > start:
            css += "background: linear-gradient(90deg,"
            if start > 0:
                css += f" {color} {start:.1f}%, transparent {start:.1f}%, "
            min_e = min(end, width)
            css += f"transparent {min_e:.1f}%, {color} {min_e:.1f}%)"
        elif end == start == 0:
            css += "background: linear-gradient(90deg,"
            css += f" {color} {100:.1f}%, transparent {100:.1f}%, "
            css += f"transparent {100:.1f}%, {color} {100:.1f}%)"
        return css

    def css(row_item: float) -> str:
        if pd.isna(row_item):
            return ""

        # avoid deprecated indexing `colors[x > zero]`
        color_value = colors[1] if row_item > zero else colors[0]
        if align == "left":
            return css_bar(0, row_item, color_value)

        return css_bar(min(row_item, zero), max(row_item, zero), color_value)

    if style_object.ndim == 1:
        # print(css(normed[10]))
        return [css(x) for x in normed]

    return pd.DataFrame(
        [[css(x) for x in row] for row in normed],
        index=style_object.index,
        columns=style_object.columns,
    )


def bar_inverse(  # noqa: D417
        styler: pd.DataFrame | Styler,
        subset: Subset = None,
        axis: Axis = 0,
        color: str = "#d65f5f",
        width: float = 100,
        align: str = "left",
        min_value: float | None = None,
        max_value: float | None = None) -> Styler:
    """
    CODE MODIFIED FROM FOLLOWING.

        https://github.com/pandas-dev/pandas/blob/v1.3.2/pandas/io/formats/style.py#L2178-L2258

    Draw (inverse) bar chart in the cell backgrounds.
    Parameters
    ----------
    styler: either a pandas DataFrame or object returned by pd.DataFrame.style
    subset: label, array-like, IndexSlice, optional
        A valid 2d input to `DataFrame.loc[<subset>]`, or, in the case of a 1d input
        or single key, to `DataFrame.loc[:, <subset>]` where the columns are
        prioritised, to limit ``data`` to *before* applying the function.
    axis: {0 or 'index', 1 or 'columns', None}, default 0
        Apply to each column (``axis=0`` or ``'index'``), to each row
        (``axis=1`` or ``'columns'``), or to the entire DataFrame at once
        with ``axis=None``.
    color: str or 2-tuple/list
        If a str is passed, the color is the same for both
        negative and positive numbers. If 2-tuple/list is used, the
        first element is the color_negative and the second is the
        color_positive (eg: ['#d65f5f', '#5fba7d']).
    width: float, default 100
        A number between 0 or 100. The largest value will cover `width`
        percent of the cell's width.
    align: {'left', 'zero',' mid'}, default 'left'
        How to align the bars with the cells.
        - 'left': the min value starts at the left of the cell.
        - 'zero': a value of zero is located at the center of the cell.
        - 'mid': the center of the cell is at (max-min)/2, or
          if values are all negative (positive) the zero is aligned
          at the right (left) of the cell.
    min_value: float, optional
        Minimum bar value, defining the left hand limit
        of the bar drawing range, lower values are clipped to `min_value`.
        When None (default): the minimum value of the data will be used.
    max_value: float, optional
        Maximum bar value, defining the right hand limit
        of the bar drawing range, higher values are clipped to `max_value`.
        When None (default): the maximum value of the data will be used.
    Returns
    -------
    styler
    """  # noqa: D410, D411
    if isinstance(styler, pd.DataFrame):
        styler = styler.style

    if align not in ("left", "zero", "mid"):
        raise ValueError("`align` must be one of {'left', 'zero',' mid'}")

    if not is_list_like(color):
        color = [color, color]
    elif len(color) == 1:
        color = [color[0], color[0]]
    elif len(color) > 2:
        raise ValueError(
            "`color` must be string or a list-like "
            "of length 2: [`color_neg`, `color_pos`] "
            "(eg: color=['#d65f5f', '#5fba7d'])",
        )

    if subset is None:
        subset = styler.data.select_dtypes(include=np.number).columns

    styler.apply(
        __bar_inverse,
        subset=subset,
        axis=axis,
        align=align,
        colors=color,
        width=width,
        min_value=min_value,
        max_value=max_value,
    )

    return styler


def html_escape_dataframe(dataframe: pd.DataFrame) -> pd.DataFrame:
    """
    HTML `escapes` all string and categorical columns and indexes in the `dataframe`.

    This can be used when displaying pd.DataFrames in Jupyter notebook using `.style`;
    e.g. `<XXX>` is displayed as blank because it is not encoded.

    Args:
        dataframe: pd.DataFrame

    Returns:
        a copy of the `dataframe` with string values replaced after being html encoded via
        `html.escape()`
    """
    def __escape(value: str | None) -> str | None:
        if not any_none_nan([value]) and isinstance(value, str):
            return escape(value)
        return value

    dataframe = dataframe.copy()
    columns_to_escape = hpandas.get_string_columns(dataframe) + \
        hpandas.get_categorical_columns(dataframe)
    for column in columns_to_escape:
        dataframe[column] = dataframe[column].apply(__escape)

    if isinstance(dataframe.index, pd.MultiIndex):
        index_tuples = [tuple([__escape(x) for x in index]) for index in dataframe.index]
        dataframe.index = pd.MultiIndex.from_tuples(index_tuples)
    else:
        dataframe.index = [__escape(x) for x in dataframe.index.to_numpy()]

    if isinstance(dataframe.columns, pd.MultiIndex):
        index_tuples = [tuple([__escape(x) for x in columns]) for columns in dataframe.columns]
        dataframe.columns = pd.MultiIndex.from_tuples(index_tuples)
    else:
        dataframe.columns = [__escape(x) for x in dataframe.columns.to_numpy()]

    return dataframe



"""Contains helper functions when working with pandas objects (e.g. DataFrames, Series)."""

from __future__ import annotations
import datetime
import math
from collections.abc import Callable
from enum import Enum

import numpy as np
import pandas as pd
from pandas.api.types import is_bool_dtype, is_numeric_dtype, is_string_dtype
from pandas.io.formats.style import Styler

from helpsk.exceptions import HelpskParamValueError
import helpsk.pandas_style as pstyle
from helpsk import color


def is_series_numeric(series: pd.Series) -> bool:
    """
    Tests whether or not a pd.Series is numeric.

    NOTE: `pandas.api.types.is_numeric_dtype()` returns `True` for `bool` dtypes;
    this function will return `False` for `bool` dtypes

    Args:
        series:
            a Pandas series

    Returns:
        True if the series is numeric (explicitly boolean type is not considered
    """
    return is_numeric_dtype(series) and not is_series_bool(series)


def is_series_bool(series: pd.Series) -> bool:
    """
    Tests whether or not a pd.Series is bool.

    `is_bool_dtype(np.array([True, False, np.nan]))` evaluates to False, so we must check if any of
    the values in the series are bool

    Args:
        series:
            a Pandas series

    Returns:
        True if the series is bool
    """
    # if the series is of type bool via is_bool_dtype, let's return True
    # otherwise, we need to check the actual values since is_bool_dtype returns false if
    # the series contains None/np.nan
    if is_bool_dtype(series):
        return True

    def is_nan(value: any) -> bool:
        try:
            if np.isnan(value):
                return True
        except TypeError:
            return False

    are_booleans = [
        isinstance(x, (bool, np.bool_)) for x in series.to_numpy()
        if x is not None and not is_nan(x)
    ]

    if len(are_booleans) == 0:
        return False

    return all(are_booleans)


def is_series_date(series: pd.Series) -> bool:
    """
    Returns True if the series contains Dates or DateTimes.

    Args:
        series: a pandas Series
    Returns:
        True if the series is of type Date or DateTime
    """
    if pd.api.types.is_datetime64_any_dtype(series):
        return True

    first_valid_index = series.first_valid_index()
    if first_valid_index is not None:
        first_valid_index = series.index.tolist().index(first_valid_index)
        return isinstance(series.iloc[first_valid_index], datetime.date)

    return False


def is_series_string(series: pd.Series) -> bool:
    """
    Returns True if the series is of type string.

    Args:
        series: a pandas Series

    Returns:
        True if the series is of type string
    """
    first_valid_index = series.first_valid_index()
    if first_valid_index is None:
        return False

    first_valid_index = series.index.tolist().index(first_valid_index)

    return (is_string_dtype(series) or isinstance(series.iloc[first_valid_index], str)) \
        and not isinstance(series.dtype, pd.CategoricalDtype) \
        and not isinstance(series.iloc[first_valid_index], Enum)


def is_series_categorical(series: pd.Series) -> bool:
    """
    Returns True if the series is of type categorical.

    Args:
        series: a pandas Series

    Returns:
        True if the series is of type categorical
    """
    return isinstance(series.dtype, pd.CategoricalDtype)


def fill_na(series: pd.Series, missing_value_replacement: str = '<Missing>') -> pd.Series:
    """
    Fills missing values with `missing_value_replacement`.

    This is only necessary if `series` is a categorical object because series.fillna(...) will
    fail if `missing_value_replacement` isn't already an existing category. This function adds the
    value to the categories if it isn't already.

    Args:
        series:
            pd.Series
        missing_value_replacement:
            the replacement value
    Returns:
        pd.Series will missing values filled with `missing_value_replacement`

        NOTE: a side effect of this function is
            - for series of type categorical: the `missing_value_replacement` will be added to the
                list of categories if the value isn't already an existing category
            - for non-categorical dtypes (e.g. bool, float, etc.) the series will be converted to
                `object` so that strings and other types can coexist.
    """
    series = series.copy()

    if is_series_categorical(series):
        if missing_value_replacement not in series.cat.categories:
            series = series.cat.add_categories(missing_value_replacement)
    else:
        # if we don't do this, then if, for example, we have boolean values and we try to call
        # `.fillna(...)` then we will get `TypeError: Need to pass bool-like values.`
        series = series.astype(object)
    return series.fillna(missing_value_replacement)


def replace_all_bools_with_strings(
        series: pd.Series,
        replacements: dict | None = None) -> pd.Series:
    """
    Replaces boolean values (True/False) with string values ('True'/'False').

    Args:
        series:
            series to replace
        replacements:
            dictionary that contains keys `True` and `False` and corresponding replacement values
    Returns:
        Returns a copy of the pd.Series.
    """
    series = series.copy()

    if replacements is None:
        replacements = {True: 'True', False: 'False'}

    is_series = is_series_categorical(series)

    if is_series:
        categories = list(
            {*series.cat.categories.to_list(), replacements[True], replacements[False]},
        )
        series = pd.Series(series.to_list())

    mask = series.apply(type) != bool
    series = series.where(mask, series.replace(replacements))

    if is_series:
        series = pd.Categorical(series, categories=categories)

    return series


def relocate(
        df: pd.DataFrame,
        column: str,
        before: str | None = None,
        after: str | None = None) -> pd.DataFrame:
    """
    Relocates `column` to the position before the column specified in `before` or
    after the column specified in `after`. It returns the DataFrame with the column order adjusted
    accordingly.

    Args:
        df: DataFrame that contains columns to relocate
        column: column to relocate
        before: name of relative column; if provided, `column` will be relocated to before the
            column name specified in `before`
        after: name of relative column; if provided, `column` will be relocated to after the column
            name specified in `after`
    """
    match_column = before or after
    assert match_column

    columns_check = set(df.columns)
    columns = list(df.columns)
    assert column in columns
    assert match_column in columns
    columns.remove(column)

    insertion_index = columns.index(match_column)
    if after:
        insertion_index += 1

    columns.insert(insertion_index, column)
    assert columns_check == set(columns)
    return df[columns]


def get_numeric_columns(dataframe: pd.DataFrame) -> list[str]:
    """
    Returns the column names from the dataframe that are numeric (and not boolean).

    NOTE: `pandas.api.types.is_numeric_dtype()` returns `True` for `bool` dtypes;
    this function treats booleans as non-numeric.

    Args:
        dataframe: a pandas dataframe

    Returns:
        list of column names that correspond to numeric types. NOTE: if a column contains all
        `np.nan` values, it will count as numeric and will be returned in the list.
    """
    return [column for column in dataframe.columns if is_series_numeric(dataframe[column])]


def get_non_numeric_columns(dataframe: pd.DataFrame) -> list[str]:
    """
    Returns the column names from the dataframe that are not numeric.

    NOTE: `pandas.api.types.is_numeric_dtype()` returns `True` for `bool` dtypes;
    this function treats booleans as non-numeric.

    Returns
        list of column names that correspond to numeric types. NOTE: if a column contains all
        `np.nan` values, it will count as numeric and will not be returned in the list.
    """
    return [column for column in dataframe.columns if not is_series_numeric(dataframe[column])]


def get_date_columns(dataframe: pd.DataFrame) -> list[str]:
    """
    Returns the column names from the dataframe that are either Date or Datetime.

    Returns
        list of column names that correspond to Date or Datetime types.
    """
    return [column for column in dataframe.columns if is_series_date(dataframe[column])]


def get_string_columns(dataframe: pd.DataFrame) -> list[str]:
    """
    Returns the column names from the dataframe that are string.

    Returns
        list of column names that correspond to string types.
    """
    return [column for column in dataframe.columns if is_series_string(dataframe[column])]


def get_categorical_columns(dataframe: pd.DataFrame) -> list[str]:
    """
    Returns the column names from the dataframe that are categorical.

    Returns
        list of column names that correspond to categorical types.
    """
    return [column for column in dataframe.columns if is_series_categorical(dataframe[column])]


def reorder_categories(
        categorical: pd.Series | pd.Categorical,
        weights: pd.Series | np.ndarray | None = None,
        weight_function: Callable = np.median,
        ascending: bool = True,
        ordered: bool = False) -> pd.Categorical:
    """
    Returns copy of the `categorical` series, with categories reordered based on the number of
    occurrences of the category (if `weights` is set to None`), or (if `weights` is not None) to
    numeric values in `weights` based on the function `weight_function`.

    similar to `fct_reorder()` in R.

    Args:
        categorical: A collection of categorical values that will be turned into a Categorical
            object with ordered categories.
        weights:
            Numeric values used to reorder the categorical. Must be the same length as
            `categorical`. If `None`, then the categories will be reordered based on number of
            occurrences for each category.
        weight_function:
            Function that determines the order of the categories. The default is `np.median`,
            meaning that order of the categories in the returned Categorical object is determined
            by the median value (of the corresponding `values`) for each category.
        ascending:
            if True, categories are ordered in ascending order based on result of `weight_function`
        ordered:
            passed to `pd.Categorical` to indicate if returned object should be `ordered`.
    """
    if weights is None:
        ordered_categories = categorical.value_counts(ascending=ascending, dropna=True)
    else:
        if len(categorical) != len(weights):
            message = f'Length of `categorical` ({len(categorical)}) ' \
                      f'must match length of `values ({len(weights)})'
            raise HelpskParamValueError(message)

        if isinstance(weights, pd.Series):
            weights = weights.to_numpy()

        weights = pd.Series(weights, index=categorical)
        # for each categoric value, calculate the associated value of the categoric, based on the
        # func e.g. if func is np.median, get the median value associated with categoric value
        ordered_categories = weights.groupby(level=0).agg(weight_function)\
            .fillna(0).sort_values(ascending=ascending)

    # check that the unique values in categorical are a subset of the categories that we are
    # setting (subset because for categorical series there might be a category set on the series
    # (i.e. series.cat.categories) that doesn't have any associated values in the series)
    assert set(categorical.dropna()).issubset(set(ordered_categories.index.dropna()))

    return pd.Categorical(
        values=categorical,
        categories=list(ordered_categories.index.values),
        ordered=ordered,
    )


def top_n_categories(
        categorical: pd.Series | pd.Categorical,
        top_n: int = 5,
        other_category: str = 'Other',
        weights: pd.Series | np.ndarray | None = None,
        weight_function: Callable = np.median,
        ordered: bool = False) -> pd.Categorical:
    """
    Returns copy of `categorical` series, with the top `n` categories retained based on either
    the count of values (if `weight` is None) or based on the values in `weight` determined by
    `aggregate_function`, and all other categories converted to `other`.

    similar to `fct_lump()` in R.

    Args:
        categorical: A collection of categorical values that will be turned into a Categorical
            object with ordered categories.
        top_n:
            the number of categories to retain. All other categories (i.e. the values in the
            series) will be replaced with the value contained in `other_category`.
            Therefore, a total of `top_n + 1` categories will be return.
        other_category:
            the value given to all other categories
        weights:
            Numeric values used to reorder the categorical. Must be the same length as
            `categorical`.
        weight_function:
            Function that determines the order of the categories. The default is `np.median`,
            meaning that order of the categories in the returned Categorical object is determined
            by the median value (of the corresponding `values`) for each category.
        ordered:
            passed to `pd.Categorical` to indicate if returned object should be `ordered`.
    """
    # if there are less than (or the same amount of) unique values than top_n; then we don't have
    # to do anything
    if len(categorical.unique()) <= top_n:
        return categorical.copy()

    if weights is None:
        top_categories = categorical.value_counts(ascending=False, dropna=True).head(top_n)
        top_categories = list(top_categories.index.values)

    else:
        if len(categorical) != len(weights):
            message = f'Length of `categorical` ({len(categorical)}) ' \
                      f'must match length of `values ({len(weights)})'
            raise HelpskParamValueError(message)

        if isinstance(weights, pd.Series):
            weights = weights.to_numpy()

        weights = pd.Series(weights, index=categorical)
        # for each categoric value, calculate the associated value of the categoric, based on the
        # func e.g. if func is np.median, get the median value associated with categoric value
        top_categories = weights.groupby(level=0).agg(weight_function).fillna(0).\
            sort_values(ascending=False).head(top_n)
        top_categories = list(top_categories.index.values)

    # TODO: this needs to be clean up; Categorical is causing issues
    final_series = pd.Categorical(categorical)
    if other_category not in final_series.categories:
        final_series = final_series.add_categories(other_category)
    final_series = pd.Series(final_series)
    final_series[final_series.apply(lambda x: x not in top_categories).fillna(False)] = other_category  # noqa

    if other_category not in top_categories:
        top_categories = [*top_categories, other_category]

    return pd.Categorical(
        final_series,
        categories=top_categories,
        ordered=ordered,
    )


def convert_integer_series_to_categorical(
        series: pd.Series, mapping: dict,
        ordered: bool = False) -> pd.Series:
    """
    Converts a Series object from integers to Categorical for a given mapping of integers to
    strings.

    Args:
        series:
            a pandas series containing integers values
        mapping:
            dictionary containing the unique values of the integers in the current data as the key,
            and the categoric string as the value.

            The mapping dict is required to have each value in the series accounted for. However,
            the series does not have to have all the values in the mapping. (For example,
            converting a small sample of data that doesn't contain all possible values should not
            fail.)
        ordered:
            a boolean representing whether the resulting Categorical series will be ordered, or
            not.

    Returns:
        A pandas Categorical Series
    """
    # check that the keys in mapping contains all numbers found in the series
    unique_values = series.unique()
    unique_values = unique_values[~np.isnan(unique_values)]
    missing_keys = [x for x in unique_values if x not in mapping]
    if missing_keys:
        message = "The following value(s) were found in `series` " \
            f"but not in `mapping` keys `{missing_keys}`"
        raise HelpskParamValueError(message)

    # Pandas expects the underlying values to be a sequence starting with 0 (i.e. to be `0, 1, 2,
    # ...`), which won't always be the case (e.g. the values in the Series might start with 1, or
    # at any number, or might not be in sequence.
    # We need to first map the actual values to the sequence pandas expects.
    # e.g. if the actual values being mapped are [1, 2, 3] then the actual_to_expected_mapping will
    # be {actual: expected, ...} i.e. {1: 0, 2: 1, 3: 2}
    actual_to_expected_mapping = dict(zip(mapping.keys(), np.arange(len(mapping))))
    converted_series = pd.Series(series).map(actual_to_expected_mapping).fillna(-1)
    converted_series = pd.Categorical.from_codes(
        converted_series.astype(int),
        mapping.values(),
        ordered=ordered,
    )
    return pd.Series(converted_series)


def numeric_summary(
        dataframe: pd.DataFrame,
        round_by: int = 2,
        return_style: bool = True,
        sort_by_columns: bool = False) -> pd.DataFrame | Styler | None:
    """
    Provides a summary of basic stats for the numeric columns of a DataFrame. Each numeric
    column in `dataframe` will correspond to a row in the pd.DataFrame returned.

    Args:
        dataframe:
            a pandas dataframe
        round_by:
            the number of decimal places to round the results
        return_style:
            If True, returns a pd.DataFrame.style object. This can be used for displaying in
            Jupyter Notebook.
            If False, returns a pd.DataFrame
        sort_by_columns:
            If True, sorts the rows of the pd.DataFrame returned by the name of the columns of
            `dataframe`, alphabetically. If False, return the rows in the order of that of the
            original `dataframe` columns.

    Returns:
        Returns a pandas DataFrame with the following attributes (returned as columns) for each of
        the numeric columns in `dataframe` (which are returned as rows).

        `# of Non-Nulls`: The number of non-null values found for the given column.
        `# of Nulls`: The number of null values found for the given column.
        `% Nulls`: The percent of null values found for a given column.
        `# of Zeros`: The number of values that equal `0`, found for the given column.
        `% Zeros`: The percent of `0`s found.
        `Mean`: The `mean` of all the values for a given column.
        `St Dev.`: The `standard deviation` of all the values for a given column.
        `Coef of Var`: The `coefficient of variation (CV)`, is defined as the standard deviation
            divided by the mean, and describes the variability of the column's values relative to
            its mean.

            We can use this metric to compare the variation of two different variables (i.e.
            columns) that have different units or scales.
        `Skewness`: "unbiased skew"; utilizes `pandas` DataFrame underlying `.skew()` function
            https://pythontic.com/pandas/dataframe-computations/skew
        `Kurtosis`: "unbiased kurtosis ... using Fisher's definition of kurtosis"; utilizes
            `pandas` DataFrame underlying `.skew()` function
        `Min`: minimum value found
        `10%`: the value found at the 10th percentile of data
        `25%`: the value found at the 25th percentile of data
        `50%`: the value found at the 50th percentile of data
        `75%`: the value found at the 75th percentile of data
        `90%`: the value found at the 90th percentile of data
        `Max`: maximum value found
    """
    # if there aren't any numeric columns and the target variable is not numeric, we don't have
    # anything to display, return None
    numeric_columns = get_numeric_columns(dataframe=dataframe)

    if not numeric_columns:
        return None

    # column, number of nulls in column, percent of nulls in column
    null_data = [(column,
                  dataframe[column].isna().sum(),
                  round(dataframe[column].isna().sum() / len(dataframe), round_by))
                 for column in numeric_columns]
    columns, num_nulls, perc_nulls = zip(*null_data)

    # column, number of 0's, percent of 0's
    zeros_data = [(sum(dataframe[column] == 0),
                   round(sum(dataframe[column] == 0) / len(dataframe), round_by))
                  for column in numeric_columns]
    num_zeros, perc_zeros = zip(*zeros_data)
    results = pd.DataFrame(
        {'# of Non-Nulls': [dataframe[x].count() for x in numeric_columns],
         '# of Nulls': num_nulls,
         '% Nulls': perc_nulls,
         '# of Zeros': num_zeros,
         '% Zeros': perc_zeros,
         'Mean': [round(dataframe[x].mean(), round_by) for x in numeric_columns],
         'St Dev.': [round(dataframe[x].std(), round_by) for x in numeric_columns],
         'Coef of Var': [round(dataframe[x].std() / dataframe[x].mean(), round_by)
                         if dataframe[x].mean() != 0 else np.nan
                         for x in numeric_columns],
         'Skewness': [round(dataframe[x].skew(), round_by) for x in numeric_columns],
         'Kurtosis': [round(dataframe[x].kurt(), round_by) for x in numeric_columns],
         'Min': [round(dataframe[x].min(), round_by) for x in numeric_columns],
         '10%': [round(dataframe[x].quantile(q=0.10), round_by) for x in numeric_columns],
         '25%': [round(dataframe[x].quantile(q=0.25), round_by) for x in numeric_columns],
         '50%': [round(dataframe[x].quantile(q=0.50), round_by) for x in numeric_columns],
         '75%': [round(dataframe[x].quantile(q=0.75), round_by) for x in numeric_columns],
         '90%': [round(dataframe[x].quantile(q=0.90), round_by) for x in numeric_columns],
         'Max': [round(dataframe[x].max(), round_by) for x in numeric_columns]},
        index=columns)

    if sort_by_columns:
        results = results.sort_index()

    if return_style:
        results = pstyle.html_escape_dataframe(results)

        columns_to_format = [
            x for x in results.columns
            if x not in ['# of Non-Nulls', '# of Nulls', '% Nulls', '# of Zeros', '% Zeros']
        ]
        results = results.style. \
            format({
                '% Nulls': '{:,.1%}'.format,
                '% Zeros': '{:,.1%}'.format,
            }). \
            pipe(
                pstyle.format,
                subset=['# of Non-Nulls', '# of Nulls', '# of Zeros'],
                round_by=0,
            ). \
            pipe(pstyle.format, subset=columns_to_format, round_by=1). \
            highlight_between(
                left=0.00000001, right=math.inf, subset=['# of Nulls', '# of Zeros'],
                color=color.WARNING,
            ). \
            bar(subset=['% Nulls'], color=color.BAD, vmin=0, vmax=1). \
            bar(subset=['% Zeros'], color=color.GRAY, vmin=0, vmax=1). \
            bar(subset=['Coef of Var'], color=color.GRAY, vmin=0, vmax=1). \
            bar(subset=['Skewness'], color=color.GRAY, align='mid', vmin=-2, vmax=2)

    return results


def non_numeric_summary(dataframe: pd.DataFrame,
                        return_style: bool = True,
                        unique_freq_value_max_chars: int = 30,
                        sort_by_columns: bool = False) -> pd.DataFrame | Styler | None:
    """
    Provides a summary of basic stats for the non-numeric columns of a DataFrame. Each
    non-numeric column in `dataframe` will correspond to a row in the pd.DataFrame returned.

    Args:
        dataframe:
            a pandas dataframe
        return_style:
            If True, returns a pd.DataFrame.style object. This can be used for displaying in
            Jupyter Notebook.
            If False, returns a pd.DataFrame
        unique_freq_value_max_chars:
            the maximum number of characters to display in the `Most Freq. Value` column
            If the value is truncated, then `[...]` is appended to the value to indicate it was
            shortened
        sort_by_columns:
            If True, sorts the rows of the pd.DataFrame returned by the name of the columns of
            `dataframe`, alphabetically. If False, return the rows in the order of that of the
            original `dataframe` columns.

    Returns:
        Returns a pandas DataFrame with the following attributes (returned as columns) for each of
        the non-numeric columns in `dataframe` (which are returned as rows).

        `# of Non-Nulls`: The number of non-null values found for the given column.
        `# of Nulls`: The number of null values found for the given column.
        `% Nulls`: The percent of null values found (i.e. `nulls / (count + nulls)`) for a given
            column.
        `Most Freq. Value`: The most frequent value found for a given column.
        `# of Unique`: The number of unique values found for a given column.
        `% Unique`: The percent of unique values found (i.e. `unique` divided by the total number
            of values (null or non-null) for a given column.
    """
    # if there aren't any non-numeric columns and the target variable is numeric, we don't have
    # anything to display, return None
    non_numeric_columns = get_non_numeric_columns(dataframe=dataframe)
    if not non_numeric_columns:
        return None

    # column, number of nulls in column, percent of nulls in column
    null_data = [(column,
                  dataframe[column].isna().sum(),
                  round(dataframe[column].isna().sum() / len(dataframe), 3))
                 for column in non_numeric_columns]
    columns, num_nulls, perc_nulls = zip(*null_data)

    num_non_nulls = [dataframe[x].count() for x in non_numeric_columns]

    def get_top_value(series: pd.Series) -> str | None:
        counts = series.value_counts()
        return counts.index[0] if len(counts) > 0 else None

    def chop_string(value: str) -> str:
        if not isinstance(value, str):
            value = str(value)

        if len(value) > unique_freq_value_max_chars:
            value = value[0:unique_freq_value_max_chars] + '[...]'
        return value

    most_frequent_value = [chop_string(get_top_value(dataframe[x])) for x in non_numeric_columns]

    def get_number_of_unique_values(series: pd.Series) -> int | None:
        try:
            number_of_uniques = len(series.dropna().unique())
        except TypeError:  # this happens, for example, with lists
            number_of_uniques = len(series.dropna().apply(str).unique())
        return number_of_uniques

    num_unique = [get_number_of_unique_values(dataframe[x]) for x in non_numeric_columns]

    perc_unique = [np.nan if n_non_nulls == 0 else n_unique / n_non_nulls
                   for n_unique, n_non_nulls in zip(num_unique, num_non_nulls)]
    perc_unique = np.round(perc_unique, 3)
    results = pd.DataFrame(
        {
            '# of Non-Nulls': num_non_nulls,
            '# of Nulls': num_nulls,
            '% Nulls': perc_nulls,
            'Most Freq. Value': most_frequent_value,
            '# of Unique': num_unique,
            '% Unique': perc_unique,
        },
        index=columns)

    if sort_by_columns:
        results = results.sort_index()

    if return_style:
        results = pstyle.html_escape_dataframe(results)
        results = results.style.format({
            '% Nulls': '{:,.1%}'.format,
            '% Unique': '{:,.1%}'.format,
        })
        results = results. \
            pipe(
                pstyle.format,
                subset=['# of Non-Nulls', '# of Nulls'],
                round_by=0,
            ). \
            pipe(pstyle.format, subset=['# of Unique'], round_by=1). \
            highlight_between(
                left=0.00000001, right=math.inf, subset=['# of Nulls'], color=color.WARNING,
            ). \
            bar(subset=['% Nulls'], color=color.BAD, vmin=0, vmax=1). \
            bar(subset=['% Unique'], color=color.GRAY, vmin=0, vmax=1)

    return results


def print_dataframe(dataframe: pd.DataFrame) -> None:
    """
    Helper function that prints a DataFrame without replacing columns with `...` when there are
    many columns.

    Args:
        dataframe: pandas DataFrame to print
    """
    with pd.option_context('display.max_columns', None), \
            pd.option_context('display.width', 20000):
        print(dataframe)


def value_frequency(series: pd.Series, sort_by_frequency: bool = True) -> pd.DataFrame:
    """
    Shows the unique values and corresponding frequencies.

    Args:
        series:
            a Pandas series, either categorical or with integers.
        sort_by_frequency:
            if True then sort by frequency desc; otherwise sort by index (either numerically
            ascending if series is numeric, or alphabetically if non-ordered categoric, or by
            category if ordered categoric

    Returns:
        DataFrame with each of the values in `series` as a single row, and `Frequency` and
        `Percent` columns matching the number of occurrences and percent representation in the
        series.
    """
    results = pd.DataFrame({
        'Frequency': series.value_counts(normalize=False, dropna=False),
        'Percent': series.value_counts(normalize=True, dropna=False)},
    )

    if not sort_by_frequency:
        if series.dtype.name == 'category':
            if series.cat.ordered:
                results = results.sort_index()
            else:
                # pandas sorts non-ordered categories by whatever order the category shows up, not
                # alphabetically
                # so change the categories to be alphabetical and sort

                values = results.index.values.dropna().unique().tolist()  # noqa
                values.sort()
                results['temp'] = results.index
                results['temp'] = results['temp'].cat.reorder_categories(values, ordered=True)
                results = results.sort_values(['temp']).drop(columns='temp')
        else:
            results = results.sort_index()
    return results



"""Functions to help with calculating conversation/retension rates."""
from __future__ import annotations
import datetime
import numpy as np
import pandas as pd
from pandas.tseries.offsets import DateOffset
from helpsk.pandas import relocate


def _percent_converted(seconds_to_convert: pd.Series, period_in_seconds: int) -> float:
    return ((seconds_to_convert > 0) & (seconds_to_convert <= period_in_seconds)).mean()


def cohorted_conversion_rates(
        df: pd.DataFrame,
        base_timestamp: str,
        conversion_timestamp: str,
        cohort: str,
        intervals: list[tuple[int, str]],
        groups: str | None = None,
        current_datetime: str | None = None) -> pd.DataFrame:
    """
    Calculate the cohorted conversion rate for a given base timestamp and conversion timestamp.

    Args:
        df:
            The dataframe containing the data to calculate the conversion rate for. The data is
            expected to be 'events', meaning that each row represents a single event for a single
            user.
        base_timestamp:
            The column name for the base timestamp to use for the calculation.
        conversion_timestamp:
            The column name for the conversion timestamp to use for the calculation.
        cohort:
            The column name for the cohort (e.g. day, week, month) to use for the calculation.
        intervals:
            A list of intervals to calculate the conversion rate for in the form of (value, unit)
            which represents the number of units to calculate the conversion rate for.
        groups:
            The column name for the groups/segments to calculate the conversion rate for.
        current_datetime:
            The current datetime to use for calculating whether an interval is valid.
    """
    def to_seconds(value: int, unit: str) -> int:
        unit_to_seconds = {
            'seconds': 1,
            'minutes': 60,
            'hours': 3600,
            'days': 86400,
        }
        return value * unit_to_seconds[unit]

    if not current_datetime:
        current_datetime = datetime.datetime.utcnow()
    current_datetime = pd.to_datetime(current_datetime)

    group_by = [cohort]
    if groups:
        group_by.append(groups)

    def f(x: pd.Series) -> pd.Series:
        d = {}
        d['# of records'] = len(x)
        # the number of seconds from the last base timestamp in the cohort to the current datetime
        # if any record in the cohort has not had enough time elapse between the base_timestamp and
        # the time to convert based on the intervals, that cohort should have a nan value
        seconds_from_max_base = (current_datetime - x[base_timestamp].max()).total_seconds()
        for value, unit in intervals:
            interval_seconds = to_seconds(value, unit)
            if seconds_from_max_base >= interval_seconds:
                d[f'{value} {unit}'] = _percent_converted(
                    x['seconds_to_conversion'],
                    interval_seconds,
                )
            else:
                d[f'{value} {unit}'] = None
        return pd.Series(d)

    return (
        df
        .assign(seconds_to_conversion=lambda x: (x[conversion_timestamp] - x[base_timestamp]).dt.total_seconds())  # noqa
        .groupby(group_by)
        .apply(f)
        .reset_index()
        .sort_values(group_by, ascending=True)
    )


def _sort_intervals(durations: list[tuple[int, str]]) -> list[tuple[int, str]]:
    """Used to display the intervals/durations in the correct order on the graph."""
    unit_multipliers = {
        'days': 86400,  # 24 * 60 * 60
        'hours': 3600,  # 60 * 60
        'minutes': 60,
        'seconds': 1,
    }

    def to_seconds(duration_tuple: tuple[int, str]) -> int:
        value, unit = duration_tuple
        return value * unit_multipliers[unit]

    return sorted(durations, key=to_seconds, reverse=True)


def plot_cohorted_conversion_rates(
        df: pd.DataFrame,
        base_timestamp: str,
        conversion_timestamp: str,
        cohort: str,
        intervals: list[tuple[int, str]],
        groups: str | None = None,
        current_datetime: str | None = None,
        graph_type: str = 'bar',
        title: str | None = None,
        subtitle: str | None = None,
        x_axis_label: str | None = None,
        y_axis_label: str | None = None,
        legend_label: str | None = None,
        facet_col_wrap: int = 2,
        category_orders: dict | None = None,
        bar_mode: str = 'overlay',
        opacity: float = 0.9,
        height: int = 700,
        width: int | None = None,
        free_y_axis: bool = True) -> pd.DataFrame:
    """
    Calculate the cohorted converssion rate for a given base timestamp and conversion timestamp.

    Args:
        df:
            The dataframe containing the data to calculate the conversion rate for. The data is
            expected to be 'events', meaning that each row represents a single event for a single
            user.
        base_timestamp:
            The column name for the base timestamp to use for the calculation.
        conversion_timestamp:
            The column name for the conversion timestamp to use for the calculation.
        cohort:
            The column name for the cohort (e.g. day, week, month) to use for the calculation.
        intervals:
            A list of intervals to calculate the conversion rate for in the form of (value, unit)
            which represents the number of units to calculate the conversion rate for.
        groups:
            The column name for the groups/segments to calculate the conversion rate for.
        current_datetime:
            The current datetime to use for calculating whether an interval is valid.
        graph_type:
            The type of chart to use for the visualization. Either 'bar' or 'line'
        title:
            The title of the plot.
        subtitle:
            The subtitle of the plot.
        x_axis_label:
            The label for the x-axis.
        y_axis_label:
            The label for the y-axis.
        legend_label:
            The label for the legend.
        facet_col_wrap:
            The number of columns in the facet grid.
        category_orders:
            A dictionary of category orders for the plot.
        bar_mode:
            Valid options are `overlay` and `group`. Default is `overlay`. If `relative` is passed
            in it is converted to `overlay`.
        opacity:
            The opacity of the bars or lines.
        height:
            The height of the plot in pixels.
        width:
            The width of the plot in pixels.
        free_y_axis:
            Whether to allow the y-axis to be free for each facet.
    """
    import plotly_express as px
    conversions = cohorted_conversion_rates(
        df=df,
        base_timestamp=base_timestamp,
        conversion_timestamp=conversion_timestamp,
        cohort=cohort,
        intervals=intervals,
        groups=groups,
        current_datetime=current_datetime,
    )
    if not title and not subtitle:
        title = '<br><sub>This graph shows the cohorted conversion rates over time at various durations relative to the base timestamp.</sub>'  # noqa
    else:
        title = title or ''
        if subtitle:
            title += f'<br><sub>{subtitle}</sub>'

    columns = [f'{x} {y}' for x, y in _sort_intervals(intervals)]
    labels = {
        'value': y_axis_label or 'Conversion Rate',
        'variable': legend_label or 'Allowed Duration',
        cohort: x_axis_label or 'Cohort',
    }
    if category_orders:
        category_orders['variable'] = columns
    else:
        category_orders = {'variable': columns}
    hover_data = {
        'value': ':.2%',
        '# of records': ':,',
    }
    if bar_mode is None or bar_mode == 'relative':
        bar_mode = 'overlay'

    if graph_type == 'bar':
        fig = px.bar(
            conversions,
            x=cohort,
            y=columns,
            title=title,
            labels=labels,
            category_orders=category_orders,
            facet_col=groups,
            facet_col_wrap=facet_col_wrap,
            hover_data=hover_data,
            barmode=bar_mode,
            opacity=opacity,
            height=height,
            width=width,
        )
    elif graph_type == 'line':
        fig = px.line(
            conversions,
            x=cohort,
            y=columns,
            title=title,
            labels=labels,
            category_orders=category_orders,
            facet_col=groups,
            facet_col_wrap=facet_col_wrap,
            hover_data=hover_data,
            height=height,
            width=width,
        )
    else:
        raise ValueError(f'Invalid graph_type: {graph_type}')

    if groups and free_y_axis:
        fig.update_yaxes(matches=None)
        fig.for_each_yaxis(lambda yaxis: yaxis.update(showticklabels=True))

    fig.update_yaxes(tickformat=',.1%')
    return fig


def retention_matrix(
        df: pd.DataFrame,
        timestamp: str,
        unique_id: str,
        min_events: int = 1,
        intervals: str = 'week',
        current_datetime: str | None = None) -> pd.DataFrame:
    """
    Calculate the retention matrix for a given timestamp column and unique ID column.

    Args:
        df:
            The dataframe containing the data to calculate the retention matrix for. The data is
            expected to be 'events', meaning that each row represents a single event for a single
            user.
        timestamp:
            The column name for the timestamp to use for the calculation. The column must not
            contain any missing values.
        unique_id:
            The column name for the unique ID to use for the calculation. The column must not
            contain any missing values.
        min_events:
            The minimum number of events a user must have in order to be considered retained in
            subsequent weeks.
        intervals:
            The intervals to use for the cohort. Valid options are 'month', 'week' and 'day'.
            This value determines the interval to use for both the cohort and the event period.
            Meaning, columns returned will be '0', '1', '2', etc. for the number of intervals. If
            'months', then these numbers represent the number of months since the cohort. If 'week'
            then these numbers represent the number of weeks since the cohort. And so on.
        current_datetime:
            The current datetime to use to filter out any events that occurred after the current
            datetime. If None, the current UTC datetime is used.

    """
    # get the number of weeks between the event week and the cohort week
    is_month_interval = intervals == 'month'
    if intervals == 'month':
        num_days_in_internal = None
    elif intervals == 'week':
        num_days_in_internal = 7
    elif intervals == 'day':
        num_days_in_internal = 1
    else:
        raise ValueError(f'interval must be either "month", "week" or "day", not {intervals}')

    intervals = intervals[0].upper()

    if current_datetime is None:
        current_datetime = datetime.datetime.utcnow()

    df = df[[timestamp, unique_id]].copy()  # noqa: PD901
    assert not df.isna().any().any()
    # filter out any events that occurred after the current datetime
    df = df[df[timestamp] <= current_datetime]  # noqa: PD901
    # Step 1: Determine the cohort of each user (the day/week/month of their first event)
    df['event_period'] = df[timestamp].dt.to_period(intervals).dt.start_time
    df['cohort'] = (
        df.groupby(unique_id)[timestamp]
        .transform('min')
        .dt.to_period(intervals)
        .dt.start_time
    )
    if is_month_interval:
        # calculate the number of months between the cohort and the event
        df['period'] = (df['event_period'].dt.year - df['cohort'].dt.year) * 12 + \
            (df['event_period'].dt.month - df['cohort'].dt.month)
    else:
        df['period'] = (df['event_period'] - df['cohort']).dt.days // num_days_in_internal

    df['period'] = df['period'].astype(int)

    # Group by cohort_week, event_week and unique_id, then filter by min_visits
    user_events_by_period = (
        df
        .groupby(['cohort', 'period', unique_id])
        .size()
        .reset_index(name='event_count')
    )
    # weird bug(?) where in some cases pandas returns records with event_count == 0 (i.e.
    # users with no events in a given period which duplicates the cohort/unique_id)
    user_events_by_period = user_events_by_period.query('event_count > 0')
    assert not user_events_by_period[[unique_id, 'period']].duplicated().any()
    # for records where period == 0, the user is considered retained by definition (regardless of
    # number of events) so we don't need to filter
    retained_users = user_events_by_period[
        (user_events_by_period['event_count'] >= min_events) | (user_events_by_period['period'] == 0)  # noqa
    ]

    # Step 2: Create a crosstab of users by cohort week and event week
    cohort_matrix = pd.crosstab(
        retained_users['cohort'],
        retained_users['period'],
    )

    # Step 3: Calculate the retention rate by dividing the number of active users each week by the
    # cohort size
    cohort_size = cohort_matrix.iloc[:, 0]
    matrix = cohort_matrix.divide(cohort_size, axis=0).fillna(0)
    assert (matrix <= 1).all().all()
    for row in matrix.index:
        for col in matrix.columns:
            offset = {'months': col} if is_month_interval else {'days': col * num_days_in_internal}
            if row + DateOffset(**offset) > current_datetime:
                matrix.loc[row, col] = np.nan

    # Step 4: Reset index and column names for better readability
    matrix = matrix.reset_index()
    matrix.columns.name = None
    matrix.columns = matrix.columns.astype(str)
    matrix['# of unique ids'] = cohort_size.astype(int).to_numpy()
    matrix = relocate(matrix, column='# of unique ids', after='cohort')
    assert (matrix['0'] == 1).all().all()
    return matrix
